@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2022 Gentoo Foundation
+ * Copyright 2005-2026 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2008 Ned Ludd        - <solar@gentoo.org>
@@ -119,8 +119,39 @@ atom_explode_cat(const char *atom, const char *cat)
 
 	/* eat file name crap when given an (autocompleted) path */
 	if ((ptr = strstr(ret->CATEGORY, ".ebuild")) != NULL ||
-			(ptr = strstr(ret->CATEGORY, ".tbz2")) != NULL)
+		(ptr = strstr(ret->CATEGORY, ".tbz2")) != NULL ||
+		(ptr = strstr(ret->CATEGORY, ".xpak")) != NULL ||
+		(ptr = strstr(ret->CATEGORY, ".gpkg.tar")) != NULL)
+	{
 		*ptr = '\0';
+		if (ptr[1] == 'x' ||
+			ptr[1] == 'g')
+		{
+			bool valid = false;
+			/* probe for the build-id, it should be a number, but it can
+			 * be optional, which is making it difficult because
+			 * in something like PN-VER.gpkg.tar, VER should not be seen
+			 * as BUILDID, while a PN can also contain dashes
+			 * see below note on PMS 3.2, we do a loose check on -[0-9]
+			 * from the start */
+			while (--ptr > ret->CATEGORY &&
+				   isdigit(*ptr))
+				valid = true;
+			if (valid &&
+				*ptr == '-')
+			{
+				pv = ret->CATEGORY;
+				while ((pv = strchr(pv, '-')) != NULL) {
+					if (isdigit(pv[1]))
+						break;
+				}
+				if (pv != ptr) {
+					ret->BUILDID = atoll(&ptr[1]);
+					*ptr = '\0';
+				}
+			}
+		}
+	}
 
 	/* chip off the trailing ::REPO as needed */
 	if ((ptr = strstr(ret->CATEGORY, "::")) != NULL) {
@@ -446,11 +477,12 @@ atom_clone(depend_atom *atom)
 	}
 
 	ret->blocker = atom->blocker;
-	ret->pfx_op = atom->pfx_op;
-	ret->sfx_op = atom->pfx_op;
-	ret->PR_int = atom->PR_int;
-	ret->letter = atom->letter;
+	ret->pfx_op  = atom->pfx_op;
+	ret->sfx_op  = atom->pfx_op;
+	ret->PR_int  = atom->PR_int;
+	ret->letter  = atom->letter;
 	ret->slotdep = atom->slotdep;
+	ret->BUILDID = atom->BUILDID;
 
 	if (atom->suffixes != NULL) {
 		for (slen = 0; atom->suffixes[slen].suffix != VER_NORM; slen++)
@@ -982,10 +1014,14 @@ atom_compare_flg(const depend_atom *data, const depend_atom *query, int flags)
 		return _atom_compare_match(NEWER, pfx_op);
 
 	/* binpkg-multi-instance support */
-	if (data->BUILDID < query->BUILDID)
-		return _atom_compare_match(OLDER, pfx_op);
-	if (data->BUILDID > query->BUILDID)
-		return _atom_compare_match(NEWER, pfx_op);
+	if (data->BUILDID > 0 &&
+		query->BUILDID > 0)
+	{
+		if (data->BUILDID < query->BUILDID)
+			return _atom_compare_match(OLDER, pfx_op);
+		if (data->BUILDID > query->BUILDID)
+			return _atom_compare_match(NEWER, pfx_op);
+	}
 
 	return _atom_compare_match(EQUAL, pfx_op);
 }
