@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2025 Gentoo Foundation
+ * Copyright 2005-2026 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2008 Ned Ludd        - <solar@gentoo.org>
@@ -79,7 +79,9 @@ tree_open_int(const char *sroot, const char *tdir, bool quiet)
 
 static const char portcachedir_pms[] = "metadata/cache";
 static const char portcachedir_md5[] = "metadata/md5-cache";
+#ifdef ENABLE_GTREE
 static const char portcachedir_gt1[] = "metadata/repo.gtree.tar";
+#endif
 static const char portrepo_name[]    = "profiles/repo_name";
 tree_ctx *
 tree_open(const char *sroot, const char *portdir)
@@ -253,7 +255,10 @@ tree_close(tree_ctx *ctx)
 		ctx->cache.categories = NULL;  /* must happen before close_cat */
 
 		array_for_each(t, n, cat) {
+			int i;
 			/* ensure we cleanup all pkgs */
+			for (i = 0; i < cat->pkg_cnt; i++)
+				cat->pkg_ctxs[i]->cached = false;
 			cat->pkg_cur = 0;
 			tree_close_cat(cat);
 		}
@@ -452,7 +457,8 @@ void
 tree_close_cat(tree_cat_ctx *cat_ctx)
 {
 	if (cat_ctx->ctx->cache.categories != NULL &&
-			contains_set(cat_ctx->name, cat_ctx->ctx->cache.categories))
+		(cat_ctx->ctx->cache.all_categories ||
+		 contains_set(cat_ctx->name, cat_ctx->ctx->cache.categories)))
 		return;
 
 	/* cleanup unreturned pkgs when sorted (or cache in use) */
@@ -509,7 +515,8 @@ tree_pkg_ctx *
 tree_open_pkg(tree_cat_ctx *cat_ctx, const char *name)
 {
 	tree_pkg_ctx *pkg_ctx;
-	bool          isgpkg = false;
+	bool          isgpkg  = false;
+	depend_atom  *patom   = NULL;
 
 	if (cat_ctx->ctx->treetype == TREE_EBUILD &&
 		cat_ctx->ctx->ebuilddir_cat_ctx == cat_ctx)
@@ -521,18 +528,46 @@ tree_open_pkg(tree_cat_ctx *cat_ctx, const char *name)
 			return NULL;  /* invalid, must be some random other file */
 		*p = '\0';
 	} else if (cat_ctx->ctx->treetype == TREE_BINPKGS) {
-		size_t len = strlen(name);
-		char  *p   = (char *)&name[len - (sizeof(".gpkg.tar") - 1)];
+		size_t len   = strlen(name);
+		char  *p     = (char *)&name[len - (sizeof(".gpkg.tar") - 1)];
+		bool   isxpk = false;
+
+		/* we pre-build the atom here to pick up on the build_id in the
+		 * filename, if any */
+
 		if (len > sizeof(".gpkg.tar") - 1 &&
 			memcmp(p, ".gpkg.tar", sizeof(".gpkg.tar") - 1) == 0)
+		{
 			isgpkg = true;
+			patom  = atom_explode_cat(name, cat_ctx->name);
+			*p     = '\0';
+		}
 
 		if (!isgpkg &&
-			(len <= sizeof(".tbz2") - 1 ||
-			 (p = (char *)&name[len - (sizeof(".tbz2") - 1)]) == NULL ||
-			 memcmp(p, ".tbz2", sizeof(".tbz2") - 1) != 0))
-			return NULL;  /* invalid, like above */
-		*p = '\0';
+			(len > sizeof(".tbz2") - 1 &&
+			 (p = (char *)&name[len - (sizeof(".tbz2") - 1)]) != NULL &&
+			 (memcmp(p, ".tbz2", sizeof(".tbz2") - 1) == 0 ||
+			  memcmp(p, ".xpak", sizeof(".xpak") - 1) == 0)))
+		{
+			isxpk = true;
+			patom  = atom_explode_cat(name, cat_ctx->name);
+			*p    = '\0';
+		}
+
+		if (!isgpkg &&
+			!isxpk)
+		{
+			struct stat sb;
+
+			if (cat_ctx->ctx->ebuilddir_cat_ctx == cat_ctx)
+				return NULL;  /* invalid, like above but only as leaf */
+
+			/* only accept this one if it is a directory (PN) */
+			if (fstatat(cat_ctx->fd, name, &sb, 0) < 0)
+				return NULL;  /* nothing? */
+			if (!S_ISDIR(sb.st_mode))
+				return NULL;  /* invalid random file */
+		}
 	}
 
 	pkg_ctx                = xzalloc(sizeof(*pkg_ctx));
@@ -541,6 +576,7 @@ tree_open_pkg(tree_cat_ctx *cat_ctx, const char *name)
 	pkg_ctx->fd            = -1;
 	pkg_ctx->cat_ctx       = cat_ctx;
 	pkg_ctx->binpkg_isgpkg = isgpkg;
+	pkg_ctx->atom          = patom;
 
 	/* see if this pkg matches the query, here we can finally check
 	 * version conditions like >=, etc. */
@@ -647,58 +683,80 @@ tree_next_pkg_int(tree_cat_ctx *cat_ctx)
 tree_pkg_ctx *
 tree_next_pkg(tree_cat_ctx *cat_ctx)
 {
-	tree_ctx *ctx = cat_ctx->ctx;
-	tree_pkg_ctx *ret = NULL;
+	tree_ctx     *ctx     = cat_ctx->ctx;
+	tree_pkg_ctx *ret     = NULL;
+	bool          recurse = false;
+	bool          dofile  = false;
+
+	/* first resume an on-going recursed listing */
+	if (ctx->ebuilddir_pkg_ctx != NULL) {
+		ret = tree_next_pkg_int(ctx->ebuilddir_cat_ctx);
+		if (ret == NULL) {
+			tree_close_cat(ctx->ebuilddir_cat_ctx);
+			if (!cat_ctx->ctx->do_sort ||
+				cat_ctx->pkg_ctxs == NULL)
+				tree_close_pkg(ctx->ebuilddir_pkg_ctx);
+			ctx->ebuilddir_pkg_ctx = NULL;
+		} else {
+			ret->binpkg_ismulti = true;  /* no matter if this is ebuild */
+			return ret;
+		}
+	}
+
+	/* load next entry */
+	ret = tree_next_pkg_int(cat_ctx);
+	if (ret == NULL)
+		return ret;
 
 	if (ctx->treetype == TREE_EBUILD) {
 		/* serve *.ebuild files each as separate pkg_ctx with name set
-		 * to CAT/P like in VDB and metadata */
-		do {
-			if (ctx->ebuilddir_pkg_ctx == NULL) {
-				tree_ctx *pkgdir = ctx->ebuilddir_ctx;
+		 * to CAT/P like in VDB and metadata, this is the only supported
+		 * format, so if there's a non-directory in here, we just ignore
+		 * it, which is fine */
+		recurse = true;
+	} else if (ctx->treetype == TREE_BINPKGS) {
+		/* with FEATURES=binpkg-multi-instance we can have at worst a
+		 * mix here of files and directories, so we must handle both */
+		recurse = true;
+		dofile  = true;
+	}
 
-				if (pkgdir == NULL)
-					pkgdir = ctx->ebuilddir_ctx = xmalloc(sizeof(*pkgdir));
+	if (recurse) {
+		tree_ctx *pkgdir = ctx->ebuilddir_ctx;
 
-				ctx->ebuilddir_cat_ctx = NULL;
-				ctx->ebuilddir_pkg_ctx = tree_next_pkg_int(cat_ctx);
-				if (ctx->ebuilddir_pkg_ctx == NULL)
-					return NULL;
+		if (pkgdir == NULL)
+			pkgdir = ctx->ebuilddir_ctx = xmalloc(sizeof(*pkgdir));
 
-				memset(pkgdir, 0, sizeof(*pkgdir));
-				pkgdir->portroot_fd = -1;
-				pkgdir->tree_fd = cat_ctx->fd;
-				pkgdir->do_sort = ctx->do_sort;
-				pkgdir->repo = ctx->repo;
-				pkgdir->treetype = ctx->treetype;
-				pkgdir->subtree = ctx->subtree;
+		ctx->ebuilddir_cat_ctx = NULL;
+		ctx->ebuilddir_pkg_ctx = ret;
 
-				ctx->ebuilddir_cat_ctx =
-					tree_open_cat(pkgdir, ctx->ebuilddir_pkg_ctx->name);
+		memset(pkgdir, 0, sizeof(*pkgdir));
+		pkgdir->portroot_fd = -1;
+		pkgdir->tree_fd     = cat_ctx->fd;
+		pkgdir->do_sort     = ctx->do_sort;
+		pkgdir->repo        = ctx->repo;
+		pkgdir->treetype    = ctx->treetype;
+		pkgdir->subtree     = ctx->subtree;
 
-				/* opening might fail if what we found wasn't a
-				 * directory or something */
-				if (ctx->ebuilddir_cat_ctx == NULL) {
-					tree_close_pkg(ctx->ebuilddir_pkg_ctx);
-					ctx->ebuilddir_pkg_ctx = NULL;
-					continue;
-				}
+		ctx->ebuilddir_cat_ctx =
+			tree_open_cat(pkgdir, ctx->ebuilddir_pkg_ctx->name);
 
-				/* "zap" the pkg such that it looks like CAT/P */
-				ctx->ebuilddir_cat_ctx->name = cat_ctx->name;
-				ctx->ebuilddir_cat_ctx->ctx = ctx;
-			}
+		/* opening might fail if what we found wasn't a
+		 * directory or something */
+		if (ctx->ebuilddir_cat_ctx == NULL) {
+			ctx->ebuilddir_pkg_ctx = NULL;
+			if (dofile)
+				return ret;
+			else
+				tree_close_pkg(ret);
+		} else {
+			/* "zap" the pkg such that it looks like CAT/P */
+			ctx->ebuilddir_cat_ctx->name = cat_ctx->name;
+			ctx->ebuilddir_cat_ctx->ctx  = ctx;
+		}
 
-			ret = tree_next_pkg_int(ctx->ebuilddir_cat_ctx);
-			if (ret == NULL) {
-				tree_close_cat(ctx->ebuilddir_cat_ctx);
-				if (!cat_ctx->ctx->do_sort || cat_ctx->pkg_ctxs == NULL)
-					tree_close_pkg(ctx->ebuilddir_pkg_ctx);
-				ctx->ebuilddir_pkg_ctx = NULL;
-			}
-		} while (ret == NULL);
-	} else {
-		ret = tree_next_pkg_int(cat_ctx);
+		/* recurse to get the next step */
+		ret = tree_next_pkg(cat_ctx);
 	}
 
 	return ret;
@@ -1257,7 +1315,8 @@ tree_pkg_read_openfd_int(tree_pkg_ctx *pkg_ctx)
 			char buf[_Q_PATH_MAX];
 			snprintf(buf, sizeof(buf), "%s.%s", pkg_ctx->name,
 					 ctx->treetype == TREE_EBUILD ? "ebuild" :
-					 pkg_ctx->binpkg_isgpkg ? "gpkg.tar" : "tbz2");
+					 pkg_ctx->binpkg_isgpkg ? "gpkg.tar" :
+					 pkg_ctx->binpkg_ismulti ? "xpak" : "tbz2");
 			pkg_ctx->fd = openat(pkg_ctx->cat_ctx->fd, buf,
 								 O_RDONLY | O_CLOEXEC);
 		} else {
@@ -1317,8 +1376,8 @@ tree_pkg_read(tree_pkg_ctx *pkg_ctx)
 					size_t  flen;
 
 					if (hash_multiple_file_fd(pkg_ctx->fd,
-										  	  srcmd5, NULL, NULL, NULL,
-										  	  NULL, &flen, HASH_MD5) == 0)
+											  srcmd5, NULL, NULL, NULL,
+											  NULL, &flen, HASH_MD5) == 0)
 						pkg_ctx->fd = -1;
 
 					mdmd5 = tree_pkg_meta_get(spkg, _md5_);
@@ -1604,6 +1663,9 @@ tree_close_metadata(tree_metadata_xml *meta_ctx)
 void
 tree_close_pkg(tree_pkg_ctx *pkg_ctx)
 {
+	if (pkg_ctx->cached)
+		return;
+
 	if (pkg_ctx->fd >= 0)
 		close(pkg_ctx->fd);
 	if (pkg_ctx->atom != NULL)
@@ -1620,22 +1682,85 @@ tree_close_pkg(tree_pkg_ctx *pkg_ctx)
 	free(pkg_ctx);
 }
 
+static int
+tree_foreach_cache_populate_cb
+(
+	tree_pkg_ctx *ctx,
+	void         *priv
+)
+{
+	tree_cat_ctx  *cat_ctx;
+	tree_pkg_ctx  *pkg;
+	set           *cache   = priv;
+	tree_ctx      *tctx    = ctx->cat_ctx->ctx;
+	depend_atom   *atom    = tree_get_atom(ctx, true);
+	tree_pkg_meta *meta    = tree_pkg_read(ctx);
+
+	cat_ctx = get_set(atom->CATEGORY, cache);
+	if (cat_ctx == NULL) {
+		cat_ctx = tree_open_cat(tctx, ".");
+		if (cache != NULL)  /* for static code analysers */
+			add_set_value(atom->CATEGORY, cat_ctx, NULL, cache);
+		/* get a pointer from the set */
+		cat_ctx->name = contains_set(atom->CATEGORY, cache);
+	}
+
+	pkg = xcalloc(1, sizeof(*pkg));
+
+	/* intuitively this would feel like it could use a set, but since
+	 * we're going to sort it, we need it as array anyway, so this is
+	 * better, especially given that the tree should not be able to
+	 * produce duplicates */
+	cat_ctx->pkg_cnt++;
+	if (cat_ctx->pkg_cnt > cat_ctx->pkg_siz) {
+		cat_ctx->pkg_siz  = ((cat_ctx->pkg_cnt / 16) + 1) * 16;
+		cat_ctx->pkg_ctxs =
+			xrealloc(cat_ctx->pkg_ctxs,
+					 sizeof(*cat_ctx->pkg_ctxs) * cat_ctx->pkg_siz);
+	}
+	cat_ctx->pkg_ctxs[cat_ctx->pkg_cnt - 1] = pkg;
+	pkg->cached  = true;
+	pkg->cat_ctx = cat_ctx;
+	pkg->atom    = atom_clone(atom);
+	pkg->name    = xstrdup(pkg->atom->PF);
+	pkg->repo    = tctx->repo != NULL ? xstrdup(tctx->repo) : NULL;
+	if (meta != NULL) {
+		pkg->fd = -2;  /* don't try to read, we fill it in here */
+		if (tctx->treetype == TREE_PACKAGES ||
+			tctx->treetype == TREE_METADATA_GTREE)
+		{
+			/* need to copy, source is based on temp space in foreach */
+			pkg->meta = tree_clone_meta(meta);
+		} else if (tctx->treetype == TREE_BINPKGS) {
+			/* BINPKG case, this one is read/allocated separately from
+			 * xpak archive, so can just take it over */
+			pkg->meta = meta;
+			ctx->meta = NULL;  /* avoid double free */
+		}
+		pkg->binpkg_isgpkg = ctx->binpkg_isgpkg;
+	} else {
+		pkg->meta = NULL;
+	}
+
+	return 0;
+}
+
 static int
 tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 {
-	char *p;
-	char *q;
-	char *c;
-	char pkgname[_Q_PATH_MAX];
-	size_t len;
-	int ret = 0;
-	const depend_atom *query = ctx->query_atom;
+	char              *p;
+	char              *q;
+	char              *c;
+	char               pkgname[_Q_PATH_MAX];
+	size_t             len;
+	int                ret      = 0;
+	const depend_atom *query    = ctx->query_atom;
 
 	/* reused for every entry */
-	tree_cat_ctx *cat = NULL;
-	tree_pkg_ctx pkg;
-	tree_pkg_meta meta;
-	depend_atom *atom = NULL;
+	tree_cat_ctx      *cat      = NULL;
+	tree_pkg_ctx       pkg;
+	tree_pkg_meta      meta;
+	depend_atom       *atom     = NULL;
 
 	/* re-read the contents, this is necessary to make it possible to
 	 * call this function multiple times */
@@ -1694,26 +1819,46 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 					}
 					cat->pkg_ctxs = (tree_pkg_ctx **)atom;  /* for name */
 				}
-				if (meta.Q_BUILDID != NULL)
+				if (meta.Q_BUILDID != NULL) {
 					atom->BUILDID = atoi(meta.Q_BUILDID);
+				} else if (meta.Q_PATH != NULL) {
+					depend_atom *patom;
+					/* dreadful, but Portage apparently generates a
+					 * Packages file without BUILD_ID while it does
+					 * generate binpkgs with build-ids in them */
+					c = strrchr(meta.Q_PATH, '/');
+					patom = atom_explode(++c);
+					if (patom != NULL) {
+						atom->BUILDID = patom->BUILDID;
+						atom_implode(patom);
+					}
+				}
 				pkgnamelen = 0;
 				if (meta.Q_PATH != NULL) {
 					size_t plen = strlen(meta.Q_PATH);
 					if (plen > sizeof(".tbz2") - 1 &&
-						memcmp(meta.Q_PATH + plen - sizeof(".tbz2") - 1,
+						memcmp(meta.Q_PATH + plen - (sizeof(".tbz2") - 1),
 							   ".tbz2", sizeof(".tbz2") - 1) == 0)
 					{
 						pkgnamelen = snprintf(pkgname, sizeof(pkgname),
 											  "%s.tbz2", atom->PF);
 						pkgname[pkgnamelen - (sizeof(".tbz2") - 1)] = '\0';
-					}
-					if (plen > sizeof(".gpkg.tar") - 1 &&
-						memcmp(meta.Q_PATH + plen - sizeof(".gpkg.tar") - 1,
+					} else if (plen > sizeof(".xpak") - 1 &&
+							   memcmp(meta.Q_PATH + plen -
+									  (sizeof(".xpak") - 1),
+									  ".xpak", sizeof(".xpak") - 1) == 0)
+					{
+						pkgnamelen = snprintf(pkgname, sizeof(pkgname),
+											  "%s.xpak", atom->PF);
+						pkgname[pkgnamelen - (sizeof(".xpak") - 1)] = '\0';
+					} else if (plen > sizeof(".gpkg.tar") - 1 &&
+						memcmp(meta.Q_PATH + plen - (sizeof(".gpkg.tar") - 1),
 							   ".gpkg.tar", sizeof(".gpkg.tar") - 1) == 0)
 					{
 						pkgnamelen = snprintf(pkgname, sizeof(pkgname),
 											  "%s.gpkg.tar", atom->PF);
 						pkgname[pkgnamelen - (sizeof(".gpkg.tar") - 1)] = '\0';
+						pkg.binpkg_isgpkg = 1;
 					}
 				}
 				if (pkgnamelen == 0) {
@@ -1764,7 +1909,10 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 				atom_implode(atom);
 			atom = atom_explode(c);
 			/* pretend this entry is bogus if it doesn't match query */
-			if (query != NULL && atom_compare(atom, query) != EQUAL) {
+			if (query != NULL &&
+				atom != NULL &&
+				atom_compare(atom, query) != EQUAL)
+			{
 				atom_implode(atom);
 				atom = NULL;
 			}
@@ -1809,7 +1957,6 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 
 	/* ensure we don't free a garbage pointer */
 	ctx->repo = NULL;
-	ctx->do_sort = false;
 	ctx->cache.store[0] = '\0';
 
 	return ret;
@@ -2025,9 +2172,6 @@ tree_foreach_gtree(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 		}
 	}
 
-	/* ensure we don't free double */
-	ctx->repo = NULL;
-	ctx->do_sort = false;
 	if (ctx->cache.storesize > 0)
 		ctx->cache.store[0] = '\0';
 	free(cat);
@@ -2037,52 +2181,133 @@ tree_foreach_gtree(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 #endif
 
 int
-tree_foreach_pkg(tree_ctx *ctx, tree_pkg_cb callback, void *priv,
-		bool sort, const depend_atom *query)
+tree_foreach_pkg
+(
+	tree_ctx          *ctx,
+	tree_pkg_cb        callback,
+	void              *priv,
+	bool               sort,
+	const depend_atom *query
+)
 {
 	tree_cat_ctx *cat_ctx;
 	tree_pkg_ctx *pkg_ctx;
-	int ret;
+	int           ret;
+	tree_pkg_cb  *origcb   = callback;
+	void         *origpriv = priv;
+	bool          postsort = false;
+	bool          traverse = true;
 
 	if (ctx == NULL)
 		return EXIT_FAILURE;
 
-	ctx->do_sort = sort;
-	ctx->query_atom = query;
+	if (ctx->cache.all_categories) {
+		/* always exploit the cache if it exists (next_cat doesn't
+		 * consider it if sorting isn't requested) */
+		sort = true;
+	} else {
+		/* perform sorting post retrieval by caching first:
+		 * - binpkgs can be a combination of files and directories,
+		 *   sorting must happen on the total result to be correct
+		 * - packages is an external file that we need to read serially
+		 *   and we can only sort on top of a cache
+		 * - gtree should be fine, but since we can't assume it is
+		 *   (others may start producing them), it's like packages
+		 *   above, something we read serially not necessarily in the
+		 *   correct order */
+		if (sort &&
+			(ctx->treetype == TREE_BINPKGS ||
+			 ctx->treetype == TREE_PACKAGES ||
+			 ctx->treetype == TREE_METADATA_GTREE))
+		{
+			callback = tree_foreach_cache_populate_cb;
+			priv     = (void *)create_set();
+			postsort = true;
+		}
 
-	/* handle Packages (binpkgs index) file separately */
-	if (ctx->treetype == TREE_PACKAGES)
-		return tree_foreach_packages(ctx, callback, priv);
+		/* handle Packages (binpkgs index) file separately */
+		if (ctx->treetype == TREE_PACKAGES) {
+			traverse = false;
+			ret      = tree_foreach_packages(ctx, callback, priv);
+		}
 
 #ifdef ENABLE_GTREE
-	/* similar a gtree cache can be read sequentially in one go */
-	if (ctx->treetype == TREE_METADATA_GTREE)
-		return tree_foreach_gtree(ctx, callback, priv);
-	if (ctx->treetype == TREE_EBUILD &&
-		ctx->subtree != NULL &&
-		ctx->subtree->treetype == TREE_METADATA_GTREE)
-		return tree_foreach_gtree(ctx->subtree, callback, priv);
-#endif
-
-	ret = 0;
-	while ((cat_ctx = tree_next_cat(ctx))) {
-		while ((pkg_ctx = tree_next_pkg(cat_ctx))) {
-			ret |= callback(pkg_ctx, priv);
-			tree_close_pkg(pkg_ctx);
+		/* similar a gtree cache can be read sequentially in one go */
+		if (ctx->treetype == TREE_METADATA_GTREE) {
+			traverse = false;
+			ret      = tree_foreach_gtree(ctx, callback, priv);
 		}
-		tree_close_cat(cat_ctx);
+		if (ctx->treetype == TREE_EBUILD &&
+			ctx->subtree != NULL &&
+			ctx->subtree->treetype == TREE_METADATA_GTREE)
+		{
+			traverse = false;
+			ret      = tree_foreach_gtree(ctx->subtree, callback, priv);
+		}
+#endif
 	}
 
-	/* allow foreach to be called again on the same open tree */
-	if (ctx->do_sort)
-		scandir_free(ctx->cat_de, ctx->cat_cnt);
-	else
-		rewinddir(ctx->dir);
-	ctx->cat_de = NULL;
+	ret = 0;
+	if (traverse)
+	{
+		ctx->do_sort    = postsort ? false : sort;
+		ctx->query_atom = query;
+
+		while ((cat_ctx = tree_next_cat(ctx))) {
+			if (callback != NULL) {
+				while ((pkg_ctx = tree_next_pkg(cat_ctx))) {
+					ret |= callback(pkg_ctx, priv);
+					tree_close_pkg(pkg_ctx);
+				}
+			}
+			tree_close_cat(cat_ctx);
+		}
+
+		/* allow foreach to be called again on the same open tree */
+		if (ctx->do_sort &&
+			ctx->cat_de != NULL)
+		{
+			scandir_free(ctx->cat_de, ctx->cat_cnt);
+		} else if (ctx->dir != NULL) {
+			rewinddir(ctx->dir);
+		}
+	}
+
+	/* reset states */
+	ctx->cat_de  = NULL;
 	ctx->cat_cur = 0;
 	ctx->cat_cnt = 0;
+	ctx->dir     = NULL;
 	ctx->do_sort = false;
 
+	if (postsort) {
+		DECLARE_ARRAY(cats);
+		size_t n;
+
+		/* should never happen, but perhaps a tree implementation
+		 * populated something, then don't leak it */
+		if (ctx->cache.categories != NULL)
+			free_set(ctx->cache.categories);
+		ctx->cache.categories     = priv;
+		ctx->cache.all_categories = true;
+
+		/* loop through all categories, and sort the pkgs */
+		values_set(ctx->cache.categories, cats);
+		array_for_each(cats, n, cat_ctx) {
+			if (cat_ctx->pkg_cnt > 1) {
+				qsort(cat_ctx->pkg_ctxs, cat_ctx->pkg_cnt,
+					  sizeof(*cat_ctx->pkg_ctxs), tree_pkg_compar);
+			}
+		}
+		xarrayfree_int(cats);
+
+		/* do the final run this call was supposed to be for using the
+		 * (sorted) cache, the callback can be empty for tree_match_atom
+		 * when it wants to build a cache first */
+		if (origcb != NULL)
+			ret = tree_foreach_pkg_fast(ctx, origcb, origpriv, ctx->query_atom);
+	}
+
 	return ret;
 }
 
@@ -2196,57 +2421,6 @@ tree_get_atoms(tree_ctx *ctx, bool fullcpv, set *satoms)
 	return state.cpf;
 }
 
-static int
-tree_match_atom_cache_populate_cb(tree_pkg_ctx *ctx, void *priv)
-{
-	tree_cat_ctx  *cat_ctx;
-	tree_pkg_ctx  *pkg;
-	set           *cache   = priv;
-	tree_ctx      *tctx    = ctx->cat_ctx->ctx;
-	depend_atom   *atom    = tree_get_atom(ctx, true);
-	tree_pkg_meta *meta    = tree_pkg_read(ctx);
-
-	(void)priv;
-
-	cat_ctx = get_set(atom->CATEGORY, cache);
-	if (cat_ctx == NULL) {
-		cat_ctx = tree_open_cat(tctx, ".");
-		if (cache != NULL)  /* for static code analysers */
-			add_set_value(atom->CATEGORY, cat_ctx, NULL, cache);
-		/* get a pointer from the set */
-		cat_ctx->name = contains_set(atom->CATEGORY, cache);
-	}
-
-	pkg = xcalloc(1, sizeof(*pkg));
-
-	/* FIXME: this really could use a set */
-	cat_ctx->pkg_cnt++;
-	cat_ctx->pkg_ctxs = xrealloc(cat_ctx->pkg_ctxs,
-			sizeof(*cat_ctx->pkg_ctxs) * cat_ctx->pkg_cnt);
-	cat_ctx->pkg_ctxs[cat_ctx->pkg_cnt - 1] = pkg;
-	pkg->cat_ctx = cat_ctx;
-	pkg->atom = atom_clone(atom);
-	pkg->name = xstrdup(pkg->atom->PF);
-	pkg->repo = tctx->repo != NULL ? xstrdup(tctx->repo) : NULL;
-	if (meta != NULL) {
-		pkg->fd = -2;  /* don't try to read, we fill it in here */
-		if (tctx->treetype == TREE_PACKAGES) {
-			/* need to copy, source is based on temp space in foreach */
-			pkg->meta = tree_clone_meta(meta);
-		} else {
-			/* BINPKG case, this one is read/allocated separately from
-			 * xpak archive, so can just take it over */
-			pkg->meta = meta;
-			ctx->meta = NULL;  /* avoid double free */
-			pkg->binpkg_isgpkg = ctx->binpkg_isgpkg;
-		}
-	} else {
-		pkg->meta = NULL;
-	}
-
-	return 0;
-}
-
 static tree_match_ctx *
 tree_match_search_cat_int(
 		tree_cat_ctx      *cat_ctx,
@@ -2285,18 +2459,30 @@ tree_match_search_cat_int(
 			if (cat_ctx->ctx->treetype == TREE_PACKAGES &&
 				pkg_ctx->meta->Q_PATH != NULL)
 			{
-				/* binpkg-multi-instance has a PATH ready for us */
+				/* Packages file has a PATH ready for us */
 				snprintf(n->path, sizeof(n->path), "%s/%s",
 						 (char *)cat_ctx->ctx->path, pkg_ctx->meta->Q_PATH);
+			} else if (pkg_ctx->binpkg_ismulti) {
+				/* FEATURES=binpkg-multi-instance complicates things as
+				 * it creates an intermediate PN directory like ebuilds,
+				 * without multi-instance, it is more flat */
+				snprintf(n->path, sizeof(n->path), "%s/%s/%s/%s%s",
+						 (char *)cat_ctx->ctx->path,
+						 atom->CATEGORY, atom->PN, pkg_ctx->name,
+						 cat_ctx->ctx->treetype == TREE_EBUILD   ? ".ebuild" :
+						 (cat_ctx->ctx->treetype == TREE_BINPKGS ||
+						  cat_ctx->ctx->treetype == TREE_PACKAGES) ?
+						 (pkg_ctx->binpkg_isgpkg   ? ".gpkg.tar" : ".xpak")  :
+																   "");
 			} else {
 				snprintf(n->path, sizeof(n->path), "%s/%s/%s%s",
 						 (char *)cat_ctx->ctx->path,
-						 cat_ctx->name, pkg_ctx->name,
+						 atom->CATEGORY, pkg_ctx->name,
 						 cat_ctx->ctx->treetype == TREE_EBUILD   ? ".ebuild" :
 						 (cat_ctx->ctx->treetype == TREE_BINPKGS ||
 						  cat_ctx->ctx->treetype == TREE_PACKAGES) ?
 						 (pkg_ctx->binpkg_isgpkg   ? ".gpkg.tar" : ".tbz2")  :
-						                                           "");
+																   "");
 			}
 			if (flags & TREE_MATCH_METADATA)
 				n->meta = tree_pkg_read(pkg_ctx);
@@ -2321,53 +2507,15 @@ tree_match_atom(tree_ctx *ctx, const depend_atom *query, int flags)
 	tree_cat_ctx *cat_ctx;
 	tree_match_ctx *ret = NULL;
 
-	ctx->do_sort = true;     /* sort uses buffer, which cache relies on */
-	ctx->query_atom = NULL;  /* ensure the cache contains ALL pkgs */
-
-	if ((ctx->treetype == TREE_PACKAGES ||
-		 ctx->treetype == TREE_BINPKGS) &&
-		ctx->cache.categories == NULL)
-	{
-		set *cache;
-		DECLARE_ARRAY(cats);
-		size_t n;
-
-		/* Reading these formats requires us to traverse the tree for
-		 * each operation, so we better cache the entire thing in one
-		 * go.  Packages in addition may not store the atoms sorted
-		 * (it's a file with content after all) and since we rely on
-		 * this for latest and first match, it is important that we sort
-		 * the contents so we're sure */
-
-		cache = ctx->cache.categories;
-		if (ctx->cache.categories != NULL)
-			ctx->cache.categories = NULL;
-		else
-			cache = create_set();
-
-		tree_foreach_pkg(ctx,
-				tree_match_atom_cache_populate_cb, cache, true, NULL);
-
-		ctx->do_sort = true;  /* turn it back on after tree_foreach_pkg */
-		ctx->cache.all_categories = true;
-		ctx->cache.categories = cache;
-
-		/* loop through all categories, and sort the pkgs */
-		values_set(cache, cats);
-		array_for_each(cats, n, cat_ctx) {
-			if (cat_ctx->pkg_cnt > 1) {
-				qsort(cat_ctx->pkg_ctxs, cat_ctx->pkg_cnt,
-						sizeof(*cat_ctx->pkg_ctxs), tree_pkg_compar);
-			}
-		}
-		xarrayfree_int(cats);
-	}
+	ctx->query_atom = NULL;  /* if caching, ensure it contains ALL pkgs */
 
 	/* activate cache for future lookups, tree_match_atom relies on
 	 * cache behaviour from tree, which means all categories and
 	 * packages remain in memory until tree_close is being called */
 	if (ctx->cache.categories == NULL)
-		ctx->cache.categories = create_set();
+		tree_foreach_pkg_sorted(ctx, NULL, NULL, NULL);  /* force cache */
+
+	ctx->do_sort = true;     /* often forces/enables cache usage */
 
 	if (query->CATEGORY == NULL) {
 		tree_match_ctx *tret;
