@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2025 Gentoo Authors
+ * Copyright 2005-2026 Gentoo Authors
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2010 Ned Ludd        - <solar@gentoo.org>
@@ -31,6 +31,7 @@
 #include "move_file.h"
 #include "contents.h"
 #include "eat_file.h"
+#include "file_magic.h"
 #include "hash.h"
 #include "human_readable.h"
 #include "profile.h"
@@ -311,7 +312,7 @@ best_version(const depend_atom *atom, int mode)
 	int             r;
 
 	if (mode & BV_EBUILD) {
-		warn("BV_EBUILD not yet supported");
+		warn("installing from ebuild not yet supported");
 		return NULL;
 	}
 	if (mode == 0) {
@@ -333,7 +334,7 @@ best_version(const depend_atom *atom, int mode)
 
 	if (mode & BV_BINPKG) {
 		if (binpkg == NULL) {
-			binpkg = tree_open_binpkg("/", pkgdir);
+			binpkg = tree_open_binpkg(portroot, pkgdir);
 			if (binpkg == NULL) {
 				if (tmv != NULL)
 					tree_match_close(tmv);
@@ -1230,7 +1231,7 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 	mkdir("image", 0755);
 
 	if (mpkg->pkg->binpkg_isgpkg) {
-#ifdef HAVE_LIBARCHIVE
+#ifdef ENABLE_GPKG
 		/* unpack the whole thing to temp, dropping the pkg name dir, so
 		 * we end up with generic files in temp */
 		struct archive       *a;
@@ -1253,6 +1254,8 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 			if (fname == NULL)
 				continue;
 			fname++;
+			if (*fname == '\0')
+				continue;  /* bug #968185 */
 
 			/* drop compressor (and "tar" -- not to be misleading) for
 			 * easy access below */
@@ -1264,12 +1267,13 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 			archive_entry_set_pathname(entry, fname);
 
 			if (archive_write_header(t, entry) != ARCHIVE_OK)
-				err("failed to write: %s", archive_error_string(t));
+				err("failed to unpack from gpkg '%s': %s",
+					fname, archive_error_string(t));
 			while (archive_read_data_block(a, (const void **)&p,
 										   &size, &off) == ARCHIVE_OK)
 			{
 				if (archive_write_data_block(t, p, size, off) != ARCHIVE_OK)
-					err("failed to write %s: %s\n",
+					err("failed to write from gpkg '%s': %s\n",
 						fname, archive_error_string(t));
 			}
 			archive_write_finish_entry(t);
@@ -1305,16 +1309,19 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 			if (fname == NULL)
 				continue;
 			fname++;
+			if (*fname == '\0')
+				continue;  /* bug #968185 */
 
 			archive_entry_set_pathname(entry, fname);
 
 			if (archive_write_header(t, entry) != ARCHIVE_OK)
-				err("failed to write: %s", archive_error_string(t));
+				err("failed to unpack metadata '%s': %s",
+					fname, archive_error_string(t));
 			while (archive_read_data_block(a, (const void **)&p,
 										   &size, &off) == ARCHIVE_OK)
 			{
 				if (archive_write_data_block(t, p, size, off) != ARCHIVE_OK)
-					err("failed to write %s: %s\n",
+					err("failed to write metadata '%s': %s\n",
 						fname, archive_error_string(t));
 			}
 			archive_write_finish_entry(t);
@@ -1349,16 +1356,19 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 			if (fname == NULL)
 				continue;
 			fname++;
+			if (*fname == '\0')
+				continue;  /* bug #968185 */
 
 			archive_entry_set_pathname(entry, fname);
 
 			if (archive_write_header(t, entry) != ARCHIVE_OK)
-				err("failed to write: %s", archive_error_string(t));
+				err("failed to unpack image '%s': %s",
+					fname, archive_error_string(t));
 			while (archive_read_data_block(a, (const void **)&p,
 										   &size, &off) == ARCHIVE_OK)
 			{
 				if (archive_write_data_block(t, p, size, off) != ARCHIVE_OK)
-					err("failed to write %s: %s\n",
+					err("failed to write image '%s': %s\n",
 						fname, archive_error_string(t));
 			}
 			archive_write_finish_entry(t);
@@ -1373,8 +1383,7 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 #endif
 	} else {
 		int vdbfd;
-		unsigned char magic[257+6];
-		FILE *mfd;
+		int mfd;
 		FILE *tarpipe;
 		FILE *tbz2f;
 		unsigned char iobuf[8192];
@@ -1383,6 +1392,7 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 		size_t n;
 		size_t rd;
 		size_t wr;
+		file_magic_type fmt;
 
 		tbz2size = 0;
 		if ((vdbfd = open("vdb", O_RDONLY)) == -1)
@@ -1407,42 +1417,29 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 		 * lzo: 9-byte:  89 'L' 'Z' 'O' 0 d a 1a a at byte 0
 		 * br:  anything else */
 
-		compr = "brotli -dc"; /* default: brotli; has no magic header */
-		mfd = fopen(mpkg->path, "r");
-		if (mfd != NULL) {
-			size_t mlen = fread(magic, 1, sizeof(magic), mfd);
-			fclose(mfd);
+		mfd = open(mpkg->path, O_RDONLY);
+		fmt = file_magic_guess_fd(mfd);
+		if (mfd >= 0)
+			close(mfd);
 
-			if (mlen >= 3 && magic[0] == 'B' && magic[1] == 'Z' &&
-					magic[2] == 'h')
-			{
+		compr = "brotli -dc"; /* default: brotli; has no magic header */
+		switch (fmt) {
+			case FMAGIC_BZIP2:
 				compr = "bzip2 -dc";
-			} else if (mlen >= 2 &&
-					magic[0] == 037 && magic[1] == 0213)
-			{
+				break;
+			case FMAGIC_GZIP:
 				compr = "gzip -dc";
-			} else if (mlen >= 5 &&
-					magic[1] == '7' && magic[2] == 'z' &&
-					magic[3] == 'X' && magic[4] == 'Z')
-			{
+				break;
+			case FMAGIC_XZ:
 				compr = "xz -dc";
-			} else if (mlen == 257+6 &&
-					magic[257] == 'u' && magic[258] == 's' &&
-					magic[259] == 't' && magic[260] == 'a' &&
-					magic[261] == 'r' &&
-					(magic[262] == '\0' || magic[262] == ' '))
-			{
+				break;
+			case FMAGIC_TAR:
 				compr = "";
-			} else if (mlen >= 4 &&
-					magic[0] == 0x04 && magic[1] == 0x22 &&
-					magic[2] == 0x4D && magic[3] == 0x18)
-			{
+				break;
+			case FMAGIC_LZ4:
 				compr = "lz4 -dc";
-			} else if (mlen >= 4 &&
-					magic[0] >= 0x22 && magic[0] <= 0x28 &&
-					magic[1] == 0xB5 && magic[2] == 0x2F &&
-					magic[3] == 0xFD)
-			{
+				break;
+			case FMAGIC_ZSTD:
 				/*
 				 * --long=31 is needed to uncompress files compressed with
 				 * --long=xx where xx>27. The option is "safe" in the sense
@@ -1462,20 +1459,16 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 				/* If really tar -I would be used we would have to quote:
 				 * compr = "I \"zstd --long=31\"";
 				 * But actually we use a pipe (see below) */
-			} else if (mlen >= 4 &&
-					magic[0] == 'L' && magic[1] == 'Z' &&
-					magic[2] == 'I' && magic[3] == 'P')
-			{
+				break;
+			case FMAGIC_LZIP:
 				compr = "lzip -dc";
-			} else if (mlen >= 9 &&
-					magic[0] == 0x89 && magic[1] == 'L' &&
-					magic[2] == 'Z' && magic[3] == 'O' &&
-					magic[4] == 0x00 && magic[5] == 0x0D &&
-					magic[6] == 0x0A && magic[7] == 0x1A &&
-					magic[8] == 0x0A)
-			{
+				break;
+			case FMAGIC_LZO:
 				compr = "lzop -dc";
-			}
+				break;
+			default:
+				warn("unhandled compression type, please file a bug");
+				break;
 		}
 
 		/* extract the binary package data */
@@ -2004,7 +1997,6 @@ static void
 pkg_fetch(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 {
 	int  verifyret;
-	char buf[_Q_PATH_MAX];
 
 	/* qmerge -pv patch */
 	if (pretend) {
@@ -2030,20 +2022,28 @@ pkg_fetch(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 	}
 
 	if (access(mpkg->path, R_OK) != 0) {
+		char *p;
+
 		if (verbose)
 			printf("Fetching %s\n", atom_to_string(mpkg->atom));
 
 		/* fetch the package */
-		snprintf(buf, sizeof(buf), "%s/%s.tbz2",
-				mpkg->atom->CATEGORY, mpkg->atom->PF);
-		fetch(mpkg->pkg->cat_ctx->ctx->path, buf);
+		p = strrchr(mpkg->path, '/');
+		if (p != NULL)
+			p = strrchr(p, '/');
+		if (p != NULL) {
+			p++;
+			fetch(mpkg->pkg->cat_ctx->ctx->path, p);
+		} else {
+			warn("invalid path: %s, skipping", mpkg->path);
+		}
 
 		/* verify the pkg exists now. unlink if zero bytes */
 		unlink_empty(mpkg->path);
 	}
 
 	if (access(mpkg->path, R_OK) != 0) {
-		warn("Failed to fetch %s.tbz2 from %s", mpkg->atom->PF, binhost);
+		warn("Failed to fetch %s from %s", mpkg->atom->PF, binhost);
 		fflush(stderr);
 		return;
 	}
