@@ -1,8 +1,8 @@
 /*
- * Copyright 2005-2025 Gentoo Foundation
+ * Copyright 2005-2026 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
- * Copyright 2005-2010 Ned Ludd	       - <solar@gentoo.org>
+ * Copyright 2005-2010 Ned Ludd		   - <solar@gentoo.org>
  * Copyright 2005-2014 Mike Frysinger  - <vapier@gentoo.org>
  * Copyright 2018-     Fabian Groffen  - <grobian@gentoo.org>
  */
@@ -37,11 +37,12 @@
 #include "xmkdir.h"
 #include "xpak.h"
 
-#define QPKG_FLAGS "cEgpP:" COMMON_FLAGS
+#define QPKG_FLAGS "cEgxpP:" COMMON_FLAGS
 static struct option const qpkg_long_opts[] = {
 	{"clean",    no_argument, NULL, 'c'},
 	{"eclean",   no_argument, NULL, 'E'},
 	{"gpkg",     no_argument, NULL, 'g'},
+	{"xpak",     no_argument, NULL, 'x'},
 	{"pretend",  no_argument, NULL, 'p'},
 	{"pkgdir",    a_argument, NULL, 'P'},
 	COMMON_LONG_OPTS
@@ -49,7 +50,8 @@ static struct option const qpkg_long_opts[] = {
 static const char * const qpkg_opts_help[] = {
 	"clean pkgdir of files that are not installed",
 	"clean pkgdir of files that are not in the tree anymore",
-	"build gpkg instead of tbz2 package",
+	"force building of gpkg instead of BINPKG_FORMAT",
+	"force building of tbz2/xpak instead of BINPKG_FORMAT",
 	"pretend only",
 	"alternate package directory",
 	COMMON_OPTS_HELP
@@ -58,12 +60,18 @@ static const char * const qpkg_opts_help[] = {
 
 extern char pretend;
 
-static char *qpkg_bindir = NULL;
-static int eclean = 0;
+typedef struct qpkg_cb_args {
+	char     *bindir;
+	tree_ctx *binpkg;
+	tree_ctx *vdb;
+	int       clean_notintree:1;
+	int       build_gpkg:1;
+	size_t    pkgs_made;
+} qpkg_cb_args;
 
 /* figure out what dirs we want to process for cleaning and display results. */
 static int
-qpkg_clean(char *dirp)
+qpkg_clean(qpkg_cb_args *args)
 {
 	size_t n;
 	size_t disp_units = 0;
@@ -74,11 +82,10 @@ qpkg_clean(char *dirp)
 	tree_ctx *t;
 	tree_ctx *pkgs;
 	char *binatomstr;
-	depend_atom *atom;
 	char buf[_Q_PATH_MAX];
 	struct stat st;
 
-	pkgs = tree_open_binpkg(portroot, dirp);
+	pkgs = args->binpkg;
 	if (pkgs == NULL)
 		return 1;
 
@@ -87,7 +94,7 @@ qpkg_clean(char *dirp)
 		return 1;
 	array_set(bin_pkgs, bins);
 
-	if (eclean) {
+	if (args->clean_notintree) {
 		const char *overlay;
 
 		array_for_each(overlays, n, overlay) {
@@ -98,11 +105,9 @@ qpkg_clean(char *dirp)
 			}
 		}
 	} else {
-		t = tree_open_vdb(portroot, portvdb);
-		if (t != NULL) {
+		t = args->vdb;
+		if (t != NULL)
 			known_pkgs = tree_get_atoms(t, true, known_pkgs);
-			tree_close(t);
-		}
 	}
 
 	if (known_pkgs != NULL) {
@@ -117,9 +122,9 @@ qpkg_clean(char *dirp)
 	}
 
 	array_for_each(bins, n, binatomstr) {
-		snprintf(buf, sizeof(buf), "%s/%s.tbz2", dirp, binatomstr);
-		atom = atom_explode(binatomstr);
-		if (lstat(buf, &st) != -1) {
+		depend_atom    *atom = atom_explode(binatomstr);
+		tree_match_ctx *m    = tree_match_atom(pkgs, atom, 0);
+		if (lstat(m->path, &st) != -1) {
 			if (S_ISREG(st.st_mode)) {
 				disp_units = KILOBYTE;
 				if ((st.st_size / KILOBYTE) > 1000)
@@ -129,11 +134,13 @@ qpkg_clean(char *dirp)
 						DKBLUE, GREEN,
 						make_human_readable_str(st.st_size, 1, disp_units),
 						disp_units == MEGABYTE ? "MiB" : "KiB",
-						DKBLUE, NORM, atom_format("%[CAT]/%[PF]", atom));
+						DKBLUE, NORM, atom_format("%[CAT]/%[PF]%[BUILDID]",
+												  m->atom));
 			}
 			if (!pretend)
 				unlink(buf);
 		}
+		tree_match_close(m);
 		atom_implode(atom);
 	}
 
@@ -214,6 +221,7 @@ write_hashes
 		warnp("failed to write hash data");
 }
 
+#ifdef ENABLE_GPKG
 static const char *
 qgpkg_set_compression(struct archive *a)
 {
@@ -235,11 +243,12 @@ qgpkg_set_compression(struct archive *a)
 	/* none, no filtering */
 	return "";
 }
+#endif
 
 static int
-qgpkg_make(tree_pkg_ctx *pkg)
+qgpkg_make(tree_pkg_ctx *pkg, qpkg_cb_args *args)
 {
-#ifdef HAVE_LIBARCHIVE
+#ifdef ENABLE_GPKG
 	struct archive *a;
 	struct archive_entry *entry;
 	struct stat st;
@@ -262,7 +271,7 @@ qgpkg_make(tree_pkg_ctx *pkg)
 
 	if (pretend) {
 		printf(" %s-%s %s:\n",
-				GREEN, NORM, atom_format("%[CATEGORY]%[PF]", atom));
+				GREEN, NORM, atom_format("%[CATEGORY]%[PF]%[BUILDID]", atom));
 		return 0;
 	}
 
@@ -270,7 +279,7 @@ qgpkg_make(tree_pkg_ctx *pkg)
 	if (line == NULL)
 		return -1;
 
-	snprintf(tmpdir, sizeof(tmpdir), "%s/qpkg.XXXXXX", qpkg_bindir);
+	snprintf(tmpdir, sizeof(tmpdir), "%s/qpkg.XXXXXX", args->binpkg->path);
 	mask = umask(S_IRWXG | S_IRWXO);
 	i = mkstemp(tmpdir);
 	umask(mask);
@@ -281,7 +290,8 @@ qgpkg_make(tree_pkg_ctx *pkg)
 	if (mkdir(tmpdir, 0750))
 		return -3;
 
-	printf(" %s-%s %s: ", GREEN, NORM, atom_format("%[CATEGORY]%[PF]", atom));
+	printf(" %s-%s %s: ", GREEN, NORM,
+		   atom_format("%[CATEGORY]%[PF]%[BUILDID]", atom));
 	fflush(stdout);
 
 	snprintf(buf, sizeof(buf), "%s/Manifest", tmpdir);
@@ -320,8 +330,8 @@ qgpkg_make(tree_pkg_ctx *pkg)
 	snprintf(gpkg, sizeof(gpkg), "%s/metadata.tar%s", tmpdir, filter);
 	archive_write_open_filename(a, gpkg);
 
-	snprintf(buf, sizeof(buf), "%s%s/%s/%s",
-			portroot, portvdb, atom->CATEGORY, atom->PF);
+	snprintf(buf, sizeof(buf), "%s/%s/%s",
+			args->vdb->path, atom->CATEGORY, atom->PF);
 	cnt = 0;
 	if ((dirfd = open(buf, O_RDONLY)) >= 0)
 		cnt = scandirat(dirfd, ".", &files, filter_self_parent, alphasort);
@@ -487,15 +497,19 @@ qgpkg_make(tree_pkg_ctx *pkg)
 	archive_write_free(a);
 
 	/* create dirs, if necessary */
-	snprintf(buf, sizeof(buf), "%s/%s", qpkg_bindir, atom->CATEGORY);
+	if (atom->BUILDID > 0)
+		i = snprintf(buf, sizeof(buf), "%s/%s/%s",
+					 args->binpkg->path, atom->CATEGORY, atom->PN);
+	else
+		i = snprintf(buf, sizeof(buf), "%s/%s",
+					 args->binpkg->path, atom->CATEGORY);
 	mkdir_p(buf, 0755);
 
-	/* at this point we use PF as package name (and directory above,
-	 * which should match according to GLEP-78), however, Portage seems
-	 * to use <PF>-1 or something, which is unspecified at this point
-	 * what it means, or how to use it */
-	snprintf(buf, sizeof(buf), "%s/%s/%s.gpkg.tar",
-			qpkg_bindir, atom->CATEGORY, atom->PF);
+	if (atom->BUILDID > 0)
+		snprintf(buf + i, sizeof(buf) - i, "/%s-%u.gpkg.tar",
+				 atom->PF, atom->BUILDID);
+	else
+		snprintf(buf + i, sizeof(buf) - i, "/%s.gpkg.tar", atom->PF);
 	if (rename(gpkg, buf)) {
 		warnp("could not move '%s' to '%s'", gpkg, buf);
 		return 1;
@@ -519,7 +533,7 @@ qgpkg_make(tree_pkg_ctx *pkg)
 }
 
 static int
-qpkg_make(tree_pkg_ctx *pkg)
+qpkg_make(tree_pkg_ctx *pkg, qpkg_cb_args *args)
 {
 	FILE *out;
 	FILE *fp;
@@ -538,7 +552,7 @@ qpkg_make(tree_pkg_ctx *pkg)
 
 	if (pretend) {
 		printf(" %s-%s %s:\n",
-				GREEN, NORM, atom_format("%[CATEGORY]%[PF]", atom));
+				GREEN, NORM, atom_format("%[CATEGORY]%[PF]%[BUILDID]", atom));
 		return 0;
 	}
 
@@ -546,7 +560,7 @@ qpkg_make(tree_pkg_ctx *pkg)
 	if (line == NULL)
 		return -1;
 
-	snprintf(tmpdir, sizeof(tmpdir), "%s/qpkg.XXXXXX", qpkg_bindir);
+	snprintf(tmpdir, sizeof(tmpdir), "%s/qpkg.XXXXXX", args->binpkg->path);
 	mask = umask(0077);
 	i = mkstemp(tmpdir);
 	umask(mask);
@@ -582,7 +596,7 @@ qpkg_make(tree_pkg_ctx *pkg)
 	fclose(out);
 
 	printf(" %s-%s %s: ", GREEN, NORM,
-			atom_format("%[CATEGORY]%[PF]", atom));
+			atom_format("%[CATEGORY]%[PF]%[BUILDID]", atom));
 	fflush(stdout);
 
 	snprintf(tbz2, sizeof(tbz2), "%s/bin.tbz2", tmpdir);
@@ -605,8 +619,8 @@ qpkg_make(tree_pkg_ctx *pkg)
 	}
 	xpaksize = st.st_size;
 
-	snprintf(buf, sizeof(buf), "%s%s/%s/%s",
-			portroot, portvdb, atom->CATEGORY, atom->PF);
+	snprintf(buf, sizeof(buf), "%s/%s/%s",
+			args->vdb->path, atom->CATEGORY, atom->PF);
 	xpak_argv[0] = buf;
 	xpak_argv[1] = NULL;
 	xpak_create(AT_FDCWD, tbz2, 1, xpak_argv, 1, verbose);
@@ -634,11 +648,19 @@ qpkg_make(tree_pkg_ctx *pkg)
 	unlink(filelist);
 
 	/* create dirs, if necessary */
-	snprintf(buf, sizeof(buf), "%s/%s", qpkg_bindir, atom->CATEGORY);
+	if (atom->BUILDID > 0)
+		i = snprintf(buf, sizeof(buf), "%s/%s/%s",
+					 args->binpkg->path, atom->CATEGORY, atom->PN);
+	else
+		i = snprintf(buf, sizeof(buf), "%s/%s",
+					 args->binpkg->path, atom->CATEGORY);
 	mkdir_p(buf, 0755);
 
-	snprintf(buf, sizeof(buf), "%s/%s/%s.tbz2",
-			qpkg_bindir, atom->CATEGORY, atom->PF);
+	if (atom->BUILDID > 0)
+		snprintf(buf + i, sizeof(buf) - i, "/%s-%u.xpak",
+				 atom->PF, atom->BUILDID);
+	else
+		snprintf(buf + i, sizeof(buf) - i, "/%s.tbz2", atom->PF);
 	if (rename(tbz2, buf)) {
 		warnp("could not move '%s' to '%s'", tbz2, buf);
 		return 1;
@@ -657,96 +679,120 @@ qpkg_make(tree_pkg_ctx *pkg)
 	return 0;
 }
 
-static int
-qgpkg_cb(tree_pkg_ctx *pkg, void *priv)
-{
-	size_t *pkgs_made = priv;
-
-	if (qgpkg_make(pkg) == 0)
-		(*pkgs_made)++;
-
-	return 0;
-}
-
 static int
 qpkg_cb(tree_pkg_ctx *pkg, void *priv)
 {
-	size_t *pkgs_made = priv;
+	qpkg_cb_args *args = priv;
 
-	if (qpkg_make(pkg) == 0)
-		(*pkgs_made)++;
+	/* check atoms to compute a build-id */
+	if (contains_set("binpkg-multi-instance", features)) {
+		depend_atom    *atom = tree_get_atom(pkg, false);
+		tree_match_ctx *m    = tree_match_atom(args->binpkg, atom,
+											   TREE_MATCH_FIRST);
+		if (m != NULL) {
+			atom->BUILDID = m->atom->BUILDID;
+			tree_match_close(m);
+		}
+
+		/* take the next, we should always start at 1, so either way
+		 * this is fine */
+		atom->BUILDID++;
+	}
+
+	if (args->build_gpkg) {
+		if (qgpkg_make(pkg, args) == 0)
+			args->pkgs_made++;
+	} else {
+		if (qpkg_make(pkg, args) == 0)
+			args->pkgs_made++;
+	}
 
 	return 0;
 }
 
 int qpkg_main(int argc, char **argv)
 {
-	tree_ctx *ctx;
-	tree_pkg_cb *cb_func = qpkg_cb;
 	size_t s;
-	size_t pkgs_made;
 	int i;
 	struct stat st;
 	depend_atom *atom;
 	int restrict_chmod = 0;
 	int qclean = 0;
 	int fd;
+	char bindir[_Q_PATH_MAX];
+	qpkg_cb_args cb_args;
+
+	memset(&cb_args, 0, sizeof(cb_args));
+
+	cb_args.bindir     = pkgdir;
+	cb_args.build_gpkg = strcmp(binpkg_format, "gpkg") == 0;
 
-	qpkg_bindir = pkgdir;
 	while ((i = GETOPT_LONG(QPKG, qpkg, "")) != -1) {
 		switch (i) {
-		case 'E': eclean = qclean = 1; break;
-		case 'c': qclean = 1; break;
-		case 'g': cb_func = qgpkg_cb; break;
-		case 'p': pretend = 1; break;
+		case 'E': cb_args.clean_notintree = true;  /* fall through */
+		case 'c': qclean = 1;                       break;
+		case 'g': cb_args.build_gpkg = true;        break;
+		case 'x': cb_args.build_gpkg = false;       break;
+		case 'p': pretend = 1;                      break;
 		case 'P':
 			restrict_chmod = 1;
-			qpkg_bindir = optarg;
-			if (access(qpkg_bindir, W_OK) != 0)
-				errp("%s", qpkg_bindir);
+			cb_args.bindir = optarg;
+			if (access(cb_args.bindir, W_OK) != 0)
+				errp("%s", cb_args.bindir);
 			break;
 		COMMON_GETOPTS_CASES(qpkg)
 		}
 	}
-	if (qclean)
-		return qpkg_clean(qpkg_bindir);
-
-	if (argc == optind)
-		qpkg_usage(EXIT_FAILURE);
 
 	/* setup temp dirs */
-	if (qpkg_bindir[0] != '/')
-		err("'%s' is not a valid package destination", qpkg_bindir);
+	if (cb_args.bindir[0] != '/')
+		err("'%s' is not a valid package destination", cb_args.bindir);
 	/* brute force just unlink any file or symlink, if this fails, it's
 	 * actually good :) */
-	unlink(qpkg_bindir);
-	fd = open(qpkg_bindir, O_RDONLY);
-	if ((fd == -1 && mkdir(qpkg_bindir, 0750) == -1) ||
+	snprintf(bindir, sizeof(bindir), "%s%s", portroot, cb_args.bindir);
+	unlink(bindir);
+	fd = open(bindir, O_RDONLY);
+	if ((fd == -1 && mkdir(bindir, 0750) == -1) ||
 			(fd != -1 && (fstat(fd, &st) == -1 || !S_ISDIR(st.st_mode))))
 	{
-		errp("could not create temp bindir '%s'", qpkg_bindir);
+		errp("could not create packages directory '%s'", bindir);
 	}
 	if (fd >= 0) {
 		/* fd is valid, pointing to a directory */
 		if (!restrict_chmod)
 			if (fchmod(fd, 0750) < 0)
-				errp("could not chmod(0750) temp bindir '%s'", qpkg_bindir);
+				errp("could not chmod(0750) packages directory '%s'", bindir);
 		close(fd);
 	}
+	cb_args.binpkg = tree_open_binpkg(portroot, cb_args.bindir);
+	if (cb_args.binpkg == NULL)
+		return EXIT_FAILURE;
+
+	cb_args.vdb = tree_open_vdb(portroot, portvdb);
+	if (!cb_args.vdb)
+		return EXIT_FAILURE;
+
+	if (qclean) {
+		int ret = qpkg_clean(&cb_args);
+		tree_close(cb_args.vdb);
+		tree_close(cb_args.binpkg);
+		return ret;
+	}
+
+	if (argc == optind) {
+		tree_close(cb_args.vdb);
+		tree_close(cb_args.binpkg);
+		qpkg_usage(EXIT_FAILURE);
+	}
 
 	/* we have to change to the root so that we can feed the full paths
 	 * to tar when we create the binary package. */
 	xchdir(portroot);
 
-	ctx = tree_open_vdb(portroot, portvdb);
-	if (!ctx)
-		return EXIT_FAILURE;
-
 	/* first process any arguments which point to /var/db/pkg, an
 	 * undocumented method to allow easily tab-completing into vdb as
 	 * arguments, the trailing / needs to be present for this (as tab
 	 * completion would do) */
-	pkgs_made = 0;
 	s = strlen(portvdb);
 	for (i = optind; i < argc; i++) {
 		size_t asize = strlen(argv[i]);
@@ -780,24 +826,25 @@ int qpkg_main(int argc, char **argv)
 		if (strcmp(argv[i], "world") == 0) {
 			/* this is a crude hack, we include all packages for this,
 			 * which isn't exactly @world, but all its deps too */
-			tree_foreach_pkg_fast(ctx, cb_func, &pkgs_made, NULL);
+			tree_foreach_pkg_fast(cb_args.vdb, qpkg_cb, &cb_args, NULL);
 			break;  /* no point in continuing since we did everything */
 		}
 		atom = atom_explode(argv[i]);
 		if (atom == NULL)
 			continue;
 
-		s = pkgs_made;
-		tree_foreach_pkg_fast(ctx, cb_func, &pkgs_made, atom);
-		if (s == pkgs_made)
+		s = cb_args.pkgs_made;
+		tree_foreach_pkg_fast(cb_args.vdb, qpkg_cb, &cb_args, atom);
+		if (s == cb_args.pkgs_made)
 			warn("no match for '%s'", argv[i]);
 		atom_implode(atom);
 	}
-	tree_close(ctx);
+	tree_close(cb_args.vdb);
+	tree_close(cb_args.binpkg);
 
-	if (pkgs_made)
+	if (cb_args.pkgs_made > 0)
 		qprintf(" %s*%s Packages can be found in %s\n",
-				GREEN, NORM, qpkg_bindir);
+				GREEN, NORM, cb_args.bindir);
 
-	return (pkgs_made ? EXIT_SUCCESS : EXIT_FAILURE);
+	return (cb_args.pkgs_made > 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 }
